import requests

from .constants import *
from urllib.parse import urljoin

class Api(object):
    def __init__(self, header_options, base_url=DEFAULT_SUPERGOOD_BASE_URL):
        self.base_url = base_url
        self.header_options = header_options

    def set_logger(self, logger):
        self.log = logger

    def set_event_sink_url(self, endpoint):
        self.event_sink_url = urljoin(self.base_url, endpoint)

    def set_error_sink_url(self, endpoint):
        self.error_sink_url = urljoin(self.base_url, endpoint)

    def post_events(self, payload):
        response = requests.post(self.event_sink_url, json=payload, headers=self.header_options)
        if(response.status_code == 401):
            raise Exception(ERRORS['UNAUTHORIZED'])
        if(response.status_code != 200 and response.status_code != 201):
            raise Exception(ERRORS['POSTING_EVENTS'])
        return response.json()

    def post_errors(self, data, error, message):
        json = {
            'payload': data,
            'error': str(error),
            'message': message
        }
        try:
            response = requests.post(self.error_sink_url, json=json, headers=self.header_options)
            return response.status_code
        except Exception as e:
            self.log.warning(f'Failed to report error to {self.error_sink_url}')




from requests import Session

def patch(cache_request, cache_response):
    _original_send = Session.send

    # AK: The http.client.request method takes care of caching the request
    # requests.send just wraps http.client.request
    # TODO: See if this makes sense longer term to use, instead of http.client
    def _wrap_send(_self, request, **kwargs):
        return _original_send(_self, request, **kwargs)

    Session.send = _wrap_send




import requests
import re
import pytest
from unittest.mock import patch, Mock

from pytest_httpserver import HTTPServer
from .helper import get_config
from supergood import Client
from supergood.api import Api

def test_dont_redact_single_key(httpserver: HTTPServer, mocker):
    config = get_config(included_keys=['include_me'])
    mocker.patch('supergood.api.Api.post_events', return_value=None)
    mocker.patch('supergood.api.Api.post_errors', return_value=None)
    supergood_client = Client(config=config)
    httpserver.expect_request('/200').respond_with_json({ 'string': 'abc', 'number': 123, 'include_me': 'im_included', 'float': 123.45, 'bool': True, 'complex_string': 'Alex Klarfeld 911!'})
    requests.get(httpserver.url_for('/200'))
    supergood_client.flush_cache()
    supergood_client.kill()
    args = Api.post_events.call_args[0][0]
    body = args[0]['response']['body']
    assert body['string'] == 'aaa'
    assert body['number'] == 111
    assert body['float'] == 111.11
    assert body['bool'] == False
    assert body['complex_string'] == 'Aaaa Aaaaaaaa 111*'
    assert body['include_me'] == 'im_included'



import gzip
import json
import hashlib

from pydash import get, set_
from base64 import b64encode

from .constants import DEFAULT_SUPERGOOD_BYTE_LIMIT, GZIP_START_BYTES

def hash_value(input):
    _input = input
    hash = hashlib.sha1()
    if not input:
        return ''
    if isinstance(_input, list):
        encoded = json.dumps(_input, sort_keys=True).encode()
        hash.update(encoded)
        return [b64encode(hash.digest()).decode('utf-8')]
    if isinstance(_input, dict):
        encoded = json.dumps(_input, sort_keys=True).encode()
        hash.update(encoded)
        return {'hashed': b64encode(hash.digest()).decode('utf-8')}
    if isinstance(_input, str):
        encoded = _input.encode()
        hash.update(encoded)
        return b64encode(hash.digest()).decode('utf-8')

def redact_values(input, included_keys, byte_limit=DEFAULT_SUPERGOOD_BYTE_LIMIT):
    _input = input

    payload = get(_input, 'response.body')
    payload_size = len(str(payload))

    if(payload_size >= byte_limit):
        set_(_input, 'response.body', hash_value(payload))
        return _input

    if not _input:
        return ''

    if isinstance(_input, list):
        for i, ele in enumerate(_input):
            _input[i] = redact_values(ele, included_keys, byte_limit)
    elif isinstance(_input, dict):
        for key in _input.keys():
            if key not in included_keys:
                _input[key] = redact_values(_input[key], included_keys, byte_limit)
    elif isinstance(_input, bool):
        _input = False
    elif isinstance(_input, str):
        _input = redact_string(_input)
    elif isinstance(_input, int):
        _input = int(redact_numeric(_input))
    elif isinstance(_input, float):
        _input = float(redact_numeric(_input))
    return _input

def redact_string(input):
    redacted = ""
    for i in range(len(input)):
        c = input[i]
        if c.isupper():
            redacted += "A"
        elif c.islower():
            redacted += "a"
        elif c.isnumeric():
            redacted += "1"
        elif c.isspace():
            redacted += " "
        else:
            redacted += "*"
    return redacted

def redact_numeric(input):
    redacted = ""
    _input = str(input)
    for i in range(len(_input)):
        c = _input[i]
        if c.isnumeric():
            redacted += "1"
        elif c == "-":
            redacted += "-"
        elif c == '.':
            redacted += "."
    return redacted


def safe_parse_json(input: str):
    if not input or input == '':
        return ''
    try:
        return json.loads(input)
    except Exception as e:
        return safe_decode(input)

def safe_decode(input, encoding='utf-8'):
    try:
        if isinstance(input, bytes) and input[:2] == GZIP_START_BYTES:
            return gzip.decompress(input)

        if isinstance(input, str):
            return input

        return input.decode(encoding)
    except Exception as e:
        return str(input)



import requests

from pytest_httpserver import HTTPServer
from .helper import get_config
from supergood import Client
from supergood.api import Api

TEST_BED_URL = 'http://supergood-testbed.herokuapp.com'
EXTERNAL_URL = 'https://api.ipify.org/?format=json'

def test_ignores_requests_to_ignored_domains(mocker):
    config = get_config(ignored_domains=['supergood-testbed.herokuapp.com'])
    mocker.patch('supergood.api.Api.post_events', return_value=None)
    mocker.patch('supergood.api.Api.post_errors', return_value=None)
    supergood_client = Client(config=config)
    requests.get(f'{TEST_BED_URL}/200')
    requests.get(f'{EXTERNAL_URL}')
    supergood_client.flush_cache()
    supergood_client.kill()
    args = Api.post_events.call_args[0][0]
    assert len(args) == 1
    assert EXTERNAL_URL == args[0]['request']['url']



import unittest
import sys
import requests
import time
import re
import urllib3
import aiohttp
import asyncio
import json
import pytest

from urllib.parse import urlparse
from dotenv import load_dotenv
from pytest_httpserver import HTTPServer
from werkzeug.wrappers import Response

from supergood import Client
from supergood.constants import ERRORS
from supergood.api import Api

from .helper import get_config
from .constants import *

load_dotenv()

TEST_BED_URL = 'http://supergood-testbed.herokuapp.com'

@pytest.fixture(scope='session', autouse=True)
def supergood_client(session_mocker):
    # mock_api = mocker.patch('supergood.api.Api')
    session_mocker.patch('supergood.api.Api.post_events', return_value=None).start()
    session_mocker.patch('supergood.api.Api.post_errors', return_value=None).start()

    client = Client()

    yield client

def test_captures_all_outgoing_200_http_requests(httpserver: HTTPServer, supergood_client):
    CALL_COUNT = 5

    for i in range(CALL_COUNT):
        response = requests.get(httpserver.url_for('/200'))

    supergood_client.flush_cache()

    args = Api.post_events.call_args[0][0]
    assert args[0]['request'] is not None
    assert args[0]['response'] is not None
    assert len(args) == CALL_COUNT
    supergood_client.kill()

def test_captures_non_success_status_and_errors(httpserver: HTTPServer, supergood_client):
    http_error_codes = [400, 401, 403, 404, 500, 501, 502, 503, 504]

    for code in http_error_codes:
        httpserver.expect_request(f'/{code}').respond_with_data(status=code)
        response = requests.get(httpserver.url_for(f'/{code}'))

    supergood_client.flush_cache()

    args = Api.post_events.call_args[0][0]
    assert len(args) == len(http_error_codes)

    responded_codes = set(map(lambda x: x['response']['status'], args))
    assert responded_codes == set(http_error_codes)
    supergood_client.kill()


def test_post_requests(httpserver: HTTPServer, supergood_client):
    httpserver.expect_request('/200', method='POST').respond_with_json({ 'payload': 'value' })
    response = requests.post(httpserver.url_for('/200'))
    response.raise_for_status()
    json = response.json()
    supergood_client.flush_cache()

    assert json['payload'] == 'value'
    args = Api.post_events.call_args[0][0]
    assert len(args) == 1
    supergood_client.kill()

def test_post_request_for_large_payload(httpserver: HTTPServer, supergood_client):
    response = requests.post(COMPASS_URL, headers=HEADERS, json=PAYLOAD)
    response.raise_for_status()
    json = response.json()

    supergood_client.flush_cache()

    assert json.get("mapResults", {}).get("totalItems", 0) >= 0
    args = Api.post_events.call_args[0][0][0]
    assert args['request'] is not None
    assert args['response'] is not None
    supergood_client.kill()


def test_hanging_response(httpserver: HTTPServer, supergood_client):
    HANG_TIME_IN_SECONDS = 2
    requests.get(f'{TEST_BED_URL}/200?sleep={HANG_TIME_IN_SECONDS}')
    supergood_client.flush_cache()

    args = Api.post_events.call_args[0][0]
    assert len(args) == 1
    supergood_client.kill()


def test_posting_errors(supergood_client, session_mocker):
    _mock = session_mocker.patch('supergood.api.Api.post_events', side_effect=Exception('Test Error'))
    _mock.start()

    requests.get(f'{TEST_BED_URL}/200')

    supergood_client.flush_cache()
    error_message = Api.post_errors.call_args[0][2]
    assert error_message == ERRORS['POSTING_EVENTS']

    _mock.resetmock()
    session_mocker.patch('supergood.api.Api.post_events').start()
    supergood_client.kill()


def test_different_http_library(httpserver: HTTPServer, supergood_client):
    http = urllib3.PoolManager()
    http.request('GET', f'{TEST_BED_URL}/200')

    supergood_client.flush_cache()

    args = Api.post_events.call_args[0][0]
    assert len(args) == 1
    supergood_client.kill()

def test_aiohttp_library(httpserver: HTTPServer, supergood_client):
    async def aiohttp_get_request():
        async with aiohttp.ClientSession() as session:
            async with session.get(f'{TEST_BED_URL}/200?success=true', headers=HEADERS, json=PAYLOAD) as response:
                return await response.json()

    async def aiohttp_post_request():
        async with aiohttp.ClientSession() as session:
            async with session.post(f'{TEST_BED_URL}/200?success=true', headers=HEADERS, json=PAYLOAD) as response:
                return await response.json()

    get_response = asyncio.run(aiohttp_get_request())
    post_response = asyncio.run(aiohttp_post_request())

    supergood_client.flush_cache()

    args = Api.post_events.call_args[0][0]
    assert len(args) == 2
    assert get_response['success'] == 'true'
    assert post_response['success'] == 'true'
    assert args[0]['response'] is not None
    assert args[0]['request'] is not None
    supergood_client.kill()



import aiohttp
from uuid import uuid4

from ..constants import REQUEST_ID_KEY
from ..helpers import safe_decode

def patch(cache_request, cache_response):
    _original_request = aiohttp.client.ClientSession._request
    _original_read = aiohttp.client_reqrep.ClientResponse.read

    async def _wrap_request(clientSession, method, url, *args, **kwargs):
        request_id = str(uuid4())
        body = kwargs.get('json', None) or kwargs.get('data', None)
        headers = kwargs.get('headers', None)

        cache_request(request_id, url, method, body, headers)

        response = await _original_request(clientSession, method, url, *args, **kwargs)
        setattr(response, REQUEST_ID_KEY, request_id)

        return response

    async def _wrap_read(clientResponse):
        request_id = getattr(clientResponse, REQUEST_ID_KEY)
        response_headers = clientResponse.headers
        response_status = clientResponse.status
        response_status_text = clientResponse.reason
        response_body = await _original_read(clientResponse)

        cache_response(request_id, response_body, response_headers, response_status, response_status_text)
        return response_body

    aiohttp.client.ClientSession._request = _wrap_request
    aiohttp.client_reqrep.ClientResponse.read = _wrap_read



#!/usr/bin/env python3

import http.client
import urllib3
import signal
import os
import sys
import traceback
import jsonpickle
import json
import atexit
import threading

from urllib.parse import urlparse
from uuid import uuid4
from datetime import datetime
from base64 import b64encode
from dotenv import load_dotenv
from .logger import Logger
from .api import Api
from .constants import *
from .helpers import redact_values, safe_parse_json, safe_decode

from .vendors.requests import patch as patch_requests
from .vendors.urllib3 import patch as patch_urllib3
from .vendors.http import patch as patch_http
from .vendors.aiohttp import patch as patch_aiohttp

load_dotenv()

class Client(object):

    def __init__(self,
        client_id=os.getenv('SUPERGOOD_CLIENT_ID'),
        client_secret_id=os.getenv('SUPERGOOD_CLIENT_SECRET'),
        base_url=os.getenv('SUPERGOOD_BASE_URL'),
        config={}
    ):
        self.base_url = base_url if base_url else DEFAULT_SUPERGOOD_BASE_URL

        authorization = f'{client_id}:{client_secret_id}'

        header_options = {
                'Accept' : 'application/json, text/plain, */*',
                'Content-Type' : 'application/json',
                'Authorization' : 'Basic ' + b64encode(bytes(authorization, 'utf-8')).decode('utf-8')
            }
        self.config = DEFAULT_SUPERGOOD_CONFIG
        self.config.update(config)
        self.api = Api(header_options, base_url=self.base_url)
        self.log = Logger(self.__class__.__name__, self.config, self.api)

        self.api.set_logger(self.log)
        self.api.set_event_sink_url(self.config['eventSinkEndpoint'])
        self.api.set_error_sink_url(self.config['errorSinkEndpoint'])

        # Initialize patches here
        patch_requests(self._cache_request, self._cache_response)
        patch_urllib3(self._cache_request, self._cache_response)
        patch_http(self._cache_request, self._cache_response)
        patch_aiohttp(self._cache_request, self._cache_response)
        self._request_cache = {}
        self._response_cache = {}

        self.interval = self.set_interval(self.flush_cache, self.config['flushInterval'] / 1000)

        # On clean exit, or terminated exit - exit gracefully
        atexit.register(self.close)

    def set_interval(self, func, sec):
        def func_wrapper():
            self.set_interval(func, sec)
            func()
        t = threading.Timer(sec, func_wrapper)

        # Function will exit and end when script ends
        t.daemon = True
        t.start()
        return t

    def _cache_request(self, request_id, url, method, body, headers):
        host_domain = urlparse(url).hostname
        supergood_base_url = urlparse(self.base_url).hostname
        now = datetime.now().isoformat()

        # Don't log requests supergood is making to the event database
        if (host_domain != supergood_base_url and host_domain not in self.config['ignoredDomains']):
            parsed_url = urlparse(url)
            try:
                request = {
                    'request': {
                        'id': request_id,
                        'method': method,
                        'url': url,
                        'body': redact_values(safe_parse_json(body), self.config['includedKeys']),
                        'headers': dict(headers),
                        'path': parsed_url.path,
                        'search': parsed_url.query,
                        'requestedAt': now,
                    }
                }
                self._request_cache[request_id] = request
            except Exception as e:
                exc_info = sys.exc_info()
                error_string = ''.join(traceback.format_exception(*exc_info))
                self.log.error(
                    {
                        'request': jsonpickle.encode(request, unpicklable=False),
                        'config': jsonpickle.encode(self.config , unpicklable=False)
                    },
                    error_string,
                    ERRORS['CACHING_REQUEST']
                )

    def _cache_response(
        self,
        request_id,
        response_body,
        response_headers,
        response_status,
        response_status_text
    ) -> None:
        request, response = {}, {}
        try:
            request = self._request_cache.pop(request_id, None)
            decoded_body = safe_decode(response_body)
            if(request):
                response = {
                    'body': redact_values(safe_parse_json(decoded_body), self.config['includedKeys']),
                    'headers': dict(response_headers),
                    'status': response_status,
                    'statusText': response_status_text,
                    'respondedAt': datetime.now().isoformat(),
                }
                self._response_cache[request_id] = {
                    'request': request['request'],
                    'response': response
                }
        except Exception as e:
            exc_info = sys.exc_info()
            error_string = ''.join(traceback.format_exception(*exc_info))
            self.log.error(
                {
                    'request': jsonpickle.encode(request, unpicklable=False),
                    'response': jsonpickle.encode(response, unpicklable=False),
                    'config': jsonpickle.encode(self.config , unpicklable=False)
                },
                error_string,
                ERRORS['CACHING_RESPONSE']
            )

    def close(self, *args) -> None:
        self.log.debug('Cleaning up, flushing cache gracefully.')
        self.interval.cancel()
        self.flush_cache(force=True)

    def kill(self, *args) -> None:
        self.log.debug('Killing process, flushing cache forcefully.')
        self._request_cache.clear()
        self._response_cache.clear()
        self.interval.cancel()

    def flush_cache(self, force=False) -> None:
        if(not self.config):
            self.log.info('Config not loaded yet, skipping flush')
            return

        response_keys = list(self._response_cache.keys())
        request_keys = list(self._request_cache.keys())
        # If there are no responses in cache, just exit
        if(len(response_keys) == 0 and not force):
            return

        # If we're forcing a flush but there's nothing in the cache, exit here
        if(force and len(response_keys) == 0 and len(request_keys) == 0):
            return

        data = list(self._response_cache.values())

        if(force):
            data += list(self._request_cache.values())
        try:
            self.log.debug(f'Flushing {len(data)} items')
            self.api.post_events(data)
        except Exception as e:
            exc_info = sys.exc_info()
            error_string = ''.join(traceback.format_exception(*exc_info))
            self.log.error(
                {
                    'data': jsonpickle.encode(data, unpicklable=False),
                    'config': jsonpickle.encode(self.config, unpicklable=False),
                }, error_string, ERRORS['POSTING_EVENTS'])
        finally:
            for response_key in response_keys:
                self._response_cache.pop(response_key, None)
            if(force):
                for request_key in request_keys:
                    self._request_cache.pop(request_key, None)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='supergood',
    version='1.0.16',
    author='Alex Klarfeld',
    description='The Python client for Supergood',
    long_description=long_description,
    long_description_content_type='text/markdown',
    package=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    py_modules=['supergood'],
    package_dir={'': 'src'},
    install_requires=[
        'python-dotenv==1.0.0',
        'jsonpickle',
        'urllib3>=1.26,<2.0',
        'requests',
        'aiohttp',
        'pydash==7.0.1',
    ],
    extras_require={
        'test': [
            'requests',
            'urllib3>=1.26,<2.0',
            'pytest==7.2.1',
            'pytest_httpserver==1.0.6',
            'python-dotenv==1.0.0',
            'Werkzeug',
            'jsonpickle==3.0.1',
            'pytest-mock==3.10.0'
        ]
    },
)
